import unittest

import os
from cfn_docgen import cfn_spec

class CfnSpecTestCase(unittest.TestCase):

    base_spec_filename = "CloudFormationResourceSpecification.json"
    default_region = "ap-northeast-1"

    def setUp(self) -> None:
        os.environ["CFN_DOCGEN_AWS_REGION"] = self.default_region
        return super().setUp()

    def test_load_file_no_cache(self):
        """
        if spec file does not locally exsits, download it.
        """
        filepath = os.path.join(os.path.dirname(__file__), self.base_spec_filename)
        try:
            os.remove(filepath)
        except FileNotFoundError:
            pass

        spec = cfn_spec.CfnSpecification(
            filepath=filepath
        )

        expected_keys = set(["PropertyTypes", "ResourceSpecificationVersion", "ResourceTypes"])
        self.assertSetEqual(
            expected_keys, set(spec.spec.keys())
        )

        try:
            os.remove(filepath)
        except FileNotFoundError:
            pass

    def test_get_resource_spec(self):
        """
        get resource spec from spec
        """
        resource_name = "AWS::S3::Bucket"
        spec = cfn_spec.CfnSpecification()

        resource_spec = spec.get_resource_spec(resource_name)
        self.assertEqual(type(resource_spec), dict)
        print(resource_spec.keys())

    def test_get_property_spec(self):
        """
        get property spec of a specific resource from spec
        """
        resource_name = "AWS::S3::Bucket"
        spec = cfn_spec.CfnSpecification()

        property_spec = spec.get_property_spec(resource_name)

        self.assertEqual(56, len(property_spec))




import glob
import os
import random
import shutil
import unittest
import random
import pandas as pd
import json


from . import test_util
from cfn_docgen import main
import pandas as pd
from cfn_docgen import util
from click.testing import CliRunner
from parameterized import parameterized

TEST_DIR=os.path.dirname(__file__)



def random_sampling(l:list, n:int=50):
    new_list = []
    for _ in range(n):
        i = random.randint(0, len(l)-1)
        try:
            new_list.append(l.pop(i))
        except IndexError:
            break
    return new_list

class MainTestCase(unittest.TestCase):

    cfn_filepaths = test_util.create_template_per_resources(
        os.path.join(TEST_DIR, "resources"),
    )
    n = int(os.environ.get("SAMPLING", 0))
    if n:
        cfn_filepaths = random_sampling(cfn_filepaths, n)


    sample_filepath = os.path.join(TEST_DIR, "resources", "AWSEC2Instance.json")

    
    regions=[
        ('eu-north-1',),
        ('ap-south-1',),
        ('eu-west-3',),
        ('eu-west-2',),
        ('eu-west-1',),
        ('ap-northeast-3',),
        ('ap-northeast-2',),
        ('ap-northeast-1',),
        ('sa-east-1',),
        ('ca-central-1',),
        ('ap-southeast-1',),
        ('ap-southeast-2',),
        ('eu-central-1',),
        ('us-east-1',),
        ('us-east-2',),
        ('us-west-1',),
        ('us-west-2',),
    ]


    @classmethod
    def remove_files(cls):
        remove_filepath = glob.glob(os.path.join(TEST_DIR, "resources", "*"))
        remove_filepath = [f for f in remove_filepath if not f.endswith("json")]
        for f in remove_filepath:
            try:
                os.remove(f)
            except FileNotFoundError:
                pass

    @classmethod
    def setUpClass(cls) -> None:
        if not len(cls.cfn_filepaths):
            test_util.create_template_per_resources(os.path.join(TEST_DIR, "resources"))


        cls.remove_files()
        return super().setUpClass()

    def setUp(self) -> None:
        try:
            del os.environ["CFN_DOCGEN_AWS_REGION"]
        except KeyError:
            pass
        return super().setUp()

    def load_excel_sheet(self, filepath:str, sheet_name:str) -> pd.DataFrame:
        print(filepath, sheet_name)
        return pd.read_excel(
            filepath, sheet_name=sheet_name,
            header=1, index_col=1, 
        )

    @parameterized.expand(cfn_filepaths)
    def test_main(self, filepath:str):
        expected_filepath = os.path.splitext(filepath)[0] + ".xlsx"
        runner = CliRunner()
        result = runner.invoke(
            main.main, f"--in {filepath}"
        )
        self.assertTrue(os.path.exists(expected_filepath), result.exc_info)


    @parameterized.expand(["xlsx", "md", "csv", "html"])
    def test_main_arg_fmt(self, fmt):
        in_filepath = self.cfn_filepaths[0]
        # expected_filepath_pattern = f"{in_filepath.split('.')[0]}*.{fmt}"
        expected_filepath_pattern = os.path.splitext(in_filepath)[0] + f"*.{fmt}"


        paths = glob.glob(expected_filepath_pattern)
        for p in paths:
            try:
                os.remove(p)
            except FileNotFoundError:
                pass

        runner = CliRunner()
        result = runner.invoke(
            main.main, f"--in {in_filepath} --fmt {fmt}"
        )
        paths = glob.glob(expected_filepath_pattern)
        self.assertGreater(len(paths), 0, result.exception)


    def test_main_arg_omit(self):
        in_filepath = self.sample_filepath
        expected_filepath = os.path.splitext(in_filepath)[0] + ".xlsx"
        runner = CliRunner()
        result = runner.invoke(
            # main.main, ["--in", in_filepath]
            main.main, f"--in {in_filepath}"
        )
        full_data = self.load_excel_sheet(expected_filepath, "Resources_Property_Detail")

        result = runner.invoke(
            main.main, f"--in {in_filepath} --omit"
        )
        omitted_data = self.load_excel_sheet(expected_filepath, "Resources_Property_Detail")

        self.assertEqual(full_data.shape[1], omitted_data.shape[1])
        self.assertGreater(full_data.shape[0], omitted_data.shape[0])


    @parameterized.expand(regions)
    def test_main_arg_region(self, region):

        in_filepath = self.sample_filepath

        specs_by_region = glob.glob(os.path.join(util.CACHE_BASE_DIR, "*", "CloudFormationResourceSpecification.json"))
        for spec in specs_by_region:
            try:
                os.remove(spec)
            except FileNotFoundError as ex:
                raise ex

        runner = CliRunner()
        result = runner.invoke(
            main.main, f"--in {in_filepath} --region {region}", color=True,
        )

        specs_by_region = glob.glob(os.path.join(util.CACHE_BASE_DIR, "*", "CloudFormationResourceSpecification.json"))


        self.assertEqual(result.exit_code, 0, result.exc_info)
        self.assertEqual(len(specs_by_region), 1, result.exc_info)
        self.assertEqual(
            specs_by_region[0],
            os.path.join(util.CACHE_BASE_DIR, region, "CloudFormationResourceSpecification.json"),
            result.exc_info,
        )
    
    def test_main_arg_refresh(self):
        region = "ap-northeast-1"
        in_filepath = self.sample_filepath
        runner = CliRunner()
        result = runner.invoke(
            main.main, f"--in {in_filepath} --region {region}"
        )
        cache_filepath = os.path.join(util.CACHE_BASE_DIR, region, "CloudFormationResourceSpecification.json")
        prev_timestamp = os.path.getatime(cache_filepath)

        try:
            shutil.rmtree(os.path.join(TEST_DIR, "cache"))
        except FileNotFoundError:
            pass
        shutil.copytree(util.CACHE_BASE_DIR, os.path.join(TEST_DIR, "cache"))

        result = runner.invoke(
            main.main, f"--in {in_filepath} --region {region} --refresh"
        )
        refreshed_timestamp = os.path.getatime(cache_filepath)

        self.assertGreater(refreshed_timestamp, prev_timestamp)

        shutil.rmtree(util.CACHE_BASE_DIR)
        shutil.copytree(os.path.join(TEST_DIR, "cache"), util.CACHE_BASE_DIR, )


    def test_main_custom_resource(self):
        filepath = os.path.join(TEST_DIR, "resources", "CustomResource.json")
        with open(filepath, "w") as fp:
            fp.write(json.dumps(
                {
                    "AWSTemplateFormatVersion": "2010-09-09",
                    "Description": "Sample template using Custom Resource",
                    "Resources": {
                        "SomeCustomResource1": {
                            "Type": "Custom::SomeCustomResource",
                            "Properties": {
                                "ServiceToken": "some-token1",
                                "Region": {"Ref": "AWS::Region"},
                                "Prop1": "some-prop1",
                                "Prop2": {"KEY2": "VAL2"}
                            }
                        },
                        "SomeCustomResource2": {
                            "Type": "AWS::CloudFormation::CustomResource",
                            "Version": 1.0,
                            "Properties": {
                                "ServiceToken": "some-token2",
                                "Region": {"Ref": "AWS::Region"},
                                "Prop3": "some-prop3",
                                "Prop4": {"KEY4": "VAL4"}
                            }
                        }        
                    }
                }
            ))
        expected_props = {
            "Prop1": "some-prop1",
            "Prop2": {"KEY2": "VAL2"},
            "Prop3": "some-prop3",
            "Prop4": {"KEY4": "VAL4"}
        }
        runner = CliRunner()
        result = runner.invoke(
            main.main, f"--in {filepath} --fmt csv"
        )
        df = pd.read_csv(filepath.replace(".json", "_Resources_Property_Summary.csv"))
        self.assertEqual(df.shape[0], 2)
        df = pd.read_csv(filepath.replace(".json", "_Resources_Property_Detail.csv"))
        for k, v in expected_props.items():
            self.assertTrue(k in df["Property"].values.tolist())
            self.assertTrue(str(v) in df["Value"].values.tolist(), f"{v, df['Value'].values.tolist()}")


from collections import OrderedDict
from copy import deepcopy
import os
from typing import List
import json
import pandas as pd
import re
from cfn_flip import to_json
import tqdm
from styleframe import StyleFrame, Styler, utils

from cfn_docgen import util

logger = util.get_module_logger(__name__, util.get_verbose())


def info_logging(name:str):
    def _info_logging(fn):
        def wrap(*args, **kwargs):
            logger.info(f"start process for {name}")
            ret = fn(*args, **kwargs)
            logger.info(f"end process for {name}")
            return ret
        return wrap
    return _info_logging


class CfnResource(object):
    def __init__(self, resource_id:str, resource_def:dict) -> None:

        self.id = resource_id
        self.definition = resource_def
        self.type = resource_def["Type"]

        
        self.creation_policy_def = resource_def.get("CreationPolicy", None)
        self.creation_policy_doc = CfnTemplate.get_spec("creation_policy").get_definition(self.type)
        self.deletion_policy_def = resource_def.get("DeletionPolicy", None)
        self.deletion_policy_doc = CfnTemplate.get_spec("deletion_policy").get_definition(self.type)
        self.update_policy_def = resource_def.get("UpdatePolicy", None)
        self.update_policy_doc = CfnTemplate.get_spec("update_policy").get_definition(self.type)
        self.update_replace_polocy_def = resource_def.get("UpdateReplacePolicy", None)
        self.update_replace_polocy_doc = CfnTemplate.get_spec("update_replace_policy").get_definition(self.type)
        self.depends_on_def = resource_def.get("DependsOn", [])

        self.user_notes = resource_def.get("Metadata", {}).get("UserNotes", {})
        self.note = self.user_notes.get("ResourceNote")

        self.resource_spec = CfnTemplate.get_spec("resource_spec").get_resource_spec(self.type, self.definition)
        self.property_spec = CfnTemplate.get_spec("resource_spec").get_property_spec(self.type)

        self.properties = dict()

        self.construct_root_props()

    def construct_root_props(self):
        """
        construct root properties in definition as property object
        """
        for root_prop_id, root_prop_spec in self.resource_spec["Properties"].items():
            try:
                root_prop_def = self.definition["Properties"][root_prop_id]
            except KeyError:
                root_prop_def = None
            self.properties[root_prop_id] = CfnProperty(
                self.id, self.type, root_prop_id, root_prop_def, root_prop_spec,
                self.user_notes.get("PropNotes", None),
            )

    @classmethod
    def from_json_def(cls, prop_def:dict) -> dict:
        """
        convert json formated CfnResource into json formated template definition
        """
        prop_def = sorted(prop_def, key=lambda j: j["Property"])
        template = {
            prop_def[0]["ResourceId"] : {
                "Type": prop_def[0]["ResourceType"],
                "Metadata": {
                    "UserNotes": {
                        "ResourceNote": prop_def[0]["ResourceNote"],
                        "PropNotes": {},
                    },
                },
                "Properties": {},
            },
        }
        for prop in prop_def:
            if prop["UserNote"] is not None:
                template[prop_def[0]["ResourceId"]]["Metadata"]["UserNotes"]["PropNotes"][prop["Property"]] = prop["UserNote"]


        props = OrderedDict({p["Property"]: p["Value"] for p in prop_def if p["Value"] is not None})
        d = {}
        for key, value in props.items():
            s = d
            tokens = re.findall(r'\w+', key)
            for count, (index, next_token) in enumerate(zip(tokens, tokens[1:] + [value]), 1):
                value = next_token if count == len(tokens) else [] if next_token.isdigit() else {}
                if isinstance(s, list):
                    index = int(index)
                    while index >= len(s):
                        s.append(value)
                elif index not in s:
                    s[index] = value
                s = s[index]
        template[prop_def[0]["ResourceId"]]["Properties"] = d

        return template



    def to_df(self, name:str):
        """
        convert resource and its property as dataframe
        """
        subcategory = name.split("_")[1]

        if subcategory == "Property":
            return self.prop_to_df(subcategory)
        if subcategory == "CreationPolicy":
            return self.creation_policy_to_df(subcategory)
        if subcategory == "DeletionPolicy":
            return self.deletion_policy_to_df(subcategory)
        if subcategory == "DependsOn":
            return self.depends_on_to_df(subcategory)
        if subcategory == "UpdatePolicy":
            return self.update_policy_to_df(subcategory)
        if subcategory == "UpdateReplacePolicy":
            return self.update_replace_policy_to_df(subcategory) 

    def add_resource_level_columns(self, df:pd.DataFrame) -> pd.DataFrame:
        df.insert(0, "ResourceNote", self.note)
        df.insert(0, "ResourceType", self.type)
        df.insert(0, "ResourceId", self.id)

        return df
    
    def prop_to_df(self, name:str) -> pd.DataFrame:
        df = pd.DataFrame(columns=[
            "Property", "Value", "UserNote", "Required", "Type", "UpdateType", "Description",
        ])

        for prop_id, prop in self.properties.items():
            df = prop.set_record(df)

        df = self.add_resource_level_columns(df)

        df = df.reset_index(drop=True)
        # df = df.sort_values(["ResourceType","ResourceId", "Property"], )
        df.index.name = name

        return df

    def creation_policy_to_df(self, name:str) -> pd.DataFrame:
        if self.creation_policy_doc is None:
            return None

        records = []
        for prop_id, prop_def in self.creation_policy_doc.items():
            record = {"Property": None, "Value": None, "Description": None, "Required": None, "Type": None}
            for key, val in prop_def.items():
                if key == "Description":
                    record["Property"] = prop_id
                    record["Description"] = val
                    records.append(deepcopy(record))
                    record = {k: None for k in record.keys()}
                else:
                    for k, v in val.items():
                        record["Property"] = f"{prop_id}.{key}"
                        record[k] = v
                        if self.creation_policy_def is not None:
                            record["Value"] = self.creation_policy_def.get(prop_id, {}).get(key, None)
                    records.append(deepcopy(record))
                    record = {k: None for k in record.keys()}


        df = pd.DataFrame(records).sort_values(["Property"])

        df = self.add_resource_level_columns(df)
        
        return df

    def deletion_policy_to_df(self, name:str) -> pd.DataFrame:
        if self.deletion_policy_doc is None:
            return None

        records = []
        record = {"Property": None, "Selected": None, "Description": None}
        for key, val in self.deletion_policy_doc.items():
            record["Property"] = key
            record["Description"] = val["Description"]
            record["Selected"] = self.deletion_policy_def == key

            records.append(deepcopy(record))
            record = {k: None for k in record.keys()}

        df = pd.DataFrame(records).sort_values(["Property"])

        df = self.add_resource_level_columns(df)

        return df

    def depends_on_to_df(self, name:str) -> pd.DataFrame:
        records = []
        record = {"DependsOn": None}
        if isinstance(self.depends_on_def, str):
            record["DependsOn"] = self.depends_on_def
            records.append(deepcopy(record))
        else:
            for val in self.depends_on_def:
                record["DependsOn"] = val
                records.append(deepcopy(record))
                record = {k: None for k in record.keys()}

        df = pd.DataFrame(records)
        df = self.add_resource_level_columns(df)

        return df
        

    def update_policy_to_df(self, name:str) -> pd.DataFrame:
        if self.update_policy_doc is None:
            return None

        records = []
        for prop_id, prop_def in self.update_policy_doc.items():
            record = {"Property": None, "Value": None, "Description": None, "Required": None, "Type": None}
            for key, val in prop_def.items():
                if key == "Description":
                    record["Property"] = prop_id
                    record["Description"] = val
                else:
                    record["Property"] = f"{prop_id}.{key}"
                    for k, v in val.items():
                        record[k] = v
                    if self.update_policy_def is not None:
                        record["Value"] = self.update_policy_def.get(prop_id, {}).get(key, None)
                records.append(deepcopy(record))
                record = {k: None for k in record.keys()}

        df = pd.DataFrame(records).sort_values(["Property"])

        df = self.add_resource_level_columns(df)
        
        return df

    def update_replace_policy_to_df(self, name:str) -> pd.DataFrame:
        if self.update_replace_polocy_doc is None:
            return None

        records = []
        record = {"Property": None, "Selected": None, "Description": None}
        for key, val in self.update_replace_polocy_doc.items():
            record["Property"] = key
            record["Description"] = val["Description"]
            record["Selected"] = self.update_replace_polocy_def == key

            records.append(deepcopy(record))
            record = {k: None for k in record.keys()}

        df = pd.DataFrame(records).sort_values(["Property"])

        df = self.add_resource_level_columns(df)
        
        return df

class CfnProperty(object):

    indent_unit = 2

    def __init__(
        self, resource_id:str, resource_type:str, prop_id:str, prop_def:dict, prop_spec:dict,
        prop_notes:dict=None, parent_prop_id:str=None ,level=0, index=None
    ) -> None:

        self.resource_id = resource_id
        self.resource_type = resource_type
        self.id = prop_id
        self.definition = prop_def
        self.spec = prop_spec

        self.prop_notes = prop_notes if prop_notes is not None else dict()

        self.duplication_allowed = prop_spec.get("DuplicatesAllowed", None)
        self.type = prop_spec.get("Type", None)
        self.required = prop_spec.get("Required", None)
        self.update_type = prop_spec.get("UpdateType", None)
        self.primitive_type = prop_spec.get("PrimitiveType", None)
        self.primitive_item_type = prop_spec.get("PrimitiveItemType", None)
        self.item_type = prop_spec.get("ItemType", None)

        self.doc_description = prop_spec.get("DocDescription", None)
        self.doc_required = prop_spec.get("DocRequired", None)
        self.doc_update = prop_spec.get("DocUpdate requires", None)
        self.doc_type = prop_spec.get("DocType", None)

        self.parent_prop_id = parent_prop_id
        self.level = level
        self.index = index

        self.fqid = self.get_fqid()
        self.note = self.prop_notes.get(self.fqid, None)

        self.omitable = None


        # if this prop is primitive, set value
        if self.definition is not None and (self.primitive_type is not None or self.primitive_item_type is not None):
            self.value = self.definition
        elif self.resource_type.startswith("Custom::") or self.resource_type == "AWS::CloudFormation::CustomResource":
            self.value = self.definition
        else:
            self.value = None
        self.child_props = dict()


        self.add_child_prop()


    def get_fqid(self) -> str:
        if self.level == 0:
            return self.id
        if self.index is None:
            return self.parent_prop_id + f".{self.id}"

        return self.parent_prop_id + f"[{self.index}].{self.id}"

    def should_stop(self, resource_type:str, fqid:str) -> bool:

        # logger.error(f"{resource_type}/{fqid}")
        if self.definition is not None:
            return False

        if resource_type == "AWS::EMR::InstanceGroupConfig" and fqid == "Configurations[0].Configurations":
            return True
        if resource_type == "AWS::EMR::InstanceFleetConfig" and fqid == "InstanceTypeConfigs[0].Configurations[0].Configurations":
            return True
        if resource_type == "AWS::EMR::Cluster" and fqid == "Configurations[0].Configurations":
            return True
        if resource_type == "AWS::EMR::Cluster" and fqid == "Instances.CoreInstanceFleet.InstanceTypeConfigs[0].Configurations[0].Configurations":
            return True
        if resource_type == "AWS::EMR::Cluster" and fqid == "Instances.CoreInstanceGroup.Configurations[0].Configurations":
            return True
        if resource_type == "AWS::EMR::Cluster" and fqid == "Instances.MasterInstanceFleet.InstanceTypeConfigs[0].Configurations[0].Configurations":
            return True
        if resource_type == "AWS::EMR::Cluster" and fqid == "Instances.TaskInstanceFleets[0].InstanceTypeConfigs[0].Configurations[0].Configurations":
            return True
        if resource_type == "AWS::EMR::Cluster" and fqid == "Instances.MasterInstanceGroup.Configurations[0].Configurations":
            return True
        if resource_type == "AWS::EMR::Cluster" and fqid == "Instances.TaskInstanceGroups[0].Configurations[0].Configurations":
            return True
        if resource_type == "AWS::WAFv2::WebACL" and re.match(r"^Rules\[0\]\.Statement\.[0-9a-zA-Z]+Statement\.[0-9a-zA-Z]*(Statements\[0\]|Statement)\.", fqid):
            return True
        if resource_type == "AWS::WAFv2::RuleGroup" and re.match(r"^Rules\[0\]\.Statement\.[0-9a-zA-Z]+Statement\.[0-9a-zA-Z]*(Statements\[0\]|Statement)\.", fqid):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid == "BindingProperties.BindingProperties.Predicates[0].Or[0].Or":
            logger.debug("match")
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid == "BindingProperties.BindingProperties.Predicates[0].Or[0].And":
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid == "BindingProperties.BindingProperties.Predicates[0].And[0].And":
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid == "BindingProperties.BindingProperties.Predicates[0].And[0].Or":
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid == "Children[0].Children":
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid == "CollectionProperties.Predicate.Or[0].Or":
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid == "CollectionProperties.Predicate.Or[0].And":
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid == "CollectionProperties.Predicate.And[0].And":
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid == "CollectionProperties.Predicate.And[0].Or":
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Events.Parameters.Type.Concat[0]."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Events.Parameters.Type.Condition."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Events.Parameters.Url.Concat[0]."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Events.Parameters.Url.Condition."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Events.Parameters.Anchor.Concat[0]."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Events.Parameters.Anchor.Condition."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Events.Parameters.Target.Concat[0]."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Events.Parameters.Target.Condition."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Events.Parameters.Global.Concat[0]."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Events.Parameters.Global.Condition."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Events.Parameters.Fields.Concat[0]."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Events.Parameters.Fields.Condition."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Events.Parameters.Id.Concat[0]."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Events.Parameters.Id.Condition."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Events.Parameters.State.Set.Concat[0]."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Events.Parameters.State.Set.Condition."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Properties.Concat[0]."):
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Component" and fqid.startswith("Properties.Condition."):
            return True

        if resource_type == "AWS::AmplifyUIBuilder::Theme" and fqid == "Overrides[0].Value.Children[0].Value":
            return True
        if resource_type == "AWS::AmplifyUIBuilder::Theme" and fqid == "Values[0].Value.Children[0].Value":
            return True

        # if resource_type == "AWS::IoTWireless::TaskDefinition" and fqid == "LoRaWANUpdateGatewayTaskEntry":
        #     return True
        # if resource_type == "AWS::IoTWireless::TaskDefinition" and fqid == "LoRaWANUpdateGatewayTaskEntry":
        #     return True
        # if resource_type == "AWS::IoTWireless::TaskDefinition" and fqid == "LoRaWANUpdateGatewayTaskEntry":
        #     return True


        if resource_type == "AWS::Rekognition::StreamProcessor" and fqid == "List":
            return True

        return False

    def add_child_prop(self):
        """
        if this property has child property, set it to child attr
        child prop may be multiple(list of properties)
        """
        if self.primitive_type is not None or self.primitive_item_type is not None:
            return

        if self.should_stop(self.resource_type, self.fqid):
            return  


        prop_id = self.item_type if self.item_type is not None else self.type
        property_name = prop_id if prop_id == "Tag" else f"{self.resource_type}.{prop_id}"

        prop_spec = CfnTemplate.get_spec("resource_spec").get_property_spec(self.resource_type, property_name)


        for child_prop_id, child_prop_spec in prop_spec[property_name]["Properties"].items():
            unique_id = f"{self.resource_id}.{child_prop_id}"
        # for child_prop_id, child_prop_spec in prop_spec.items():
            if self.type is not None and self.type == "List":
                if isinstance(self.definition, list):
                    child_props = []
                    for i, prop_def in enumerate(self.definition):
                        try:
                            child_prop_def = prop_def[child_prop_id]
                        except (KeyError, TypeError):
                            child_prop_def = None

                        child_props.append(CfnProperty(
                            self.resource_id, self.resource_type, child_prop_id, child_prop_def, child_prop_spec,
                            prop_notes=self.prop_notes,
                            parent_prop_id=self.fqid,
                            level=self.level+1, index=i,
                        ))

                    self.child_props[unique_id] = deepcopy(child_props)
                else:
                    child_props = []
                    child_props.append(CfnProperty(
                        self.resource_id, self.resource_type, child_prop_id, None, child_prop_spec,
                        prop_notes=self.prop_notes,
                        parent_prop_id=self.fqid,
                        level=self.level+1, index=0,
                    ))
                    self.child_props[unique_id] = deepcopy(child_props)
            else:
                prop_def = self.definition

                try:
                    child_prop_def = prop_def[child_prop_id]
                except (KeyError, TypeError):
                    child_prop_def = None
                self.child_props[unique_id] = CfnProperty(
                    self.resource_id, self.resource_type, child_prop_id, child_prop_def, child_prop_spec,
                    prop_notes=self.prop_notes,
                    parent_prop_id=self.fqid,
                    level=self.level+1,
                )

    def repr(self) -> str:
        return self.fqid

    def is_omittable(self, is_parent_omittable:bool=False) -> bool:
        if self.omitable is not None:
            return self.omitable

        if self.value:
            self.omitable = False
            return self.omitable

        if not bool(self.child_props):
            if self.required == False:
                self.omitable = True
                return self.omitable
            else: # required = True
                self.omitable = is_parent_omittable == True
                return self.omitable

        else:
            is_current_omittable = self.required == False

        results = []
        for props in self.child_props.values():
            if isinstance(props, list):
                for prop in props:
                    results.append(prop.is_omittable(is_current_omittable))
            else:
                results.append(props.is_omittable(is_current_omittable))
        self.omitable = all(results)
        return self.omitable

    def set_record(self, df:pd.DataFrame) -> pd.DataFrame:
        tmp_df = pd.DataFrame({
            "Property": [self.repr()],
            "Value": [self.value],
            "UserNote": [self.note],
            "Required": [self.doc_required],
            "Type": [self.doc_type],
            "UpdateType": [self.doc_update],
            "IsOmittable": [self.is_omittable()],
            "Description": [self.doc_description],
            # "FQID": [self.fqid]
        })

        df = pd.concat([df, tmp_df])
        if self.value is None:
            for props in self.child_props.values():
                if isinstance(props, list):
                    for prop in props:
                        df = prop.set_record(df)
                else:
                    df = props.set_record(df)


        return df

class CfnParameter(object):

    def __init__(self, param_id:str, param_def:dict) -> None:
        self.id = param_id
        self.description = param_def.get("Description", None)
        self.allowed_pattern = param_def.get("AllowedPattern", None)
        self.allowed_values = param_def.get("AllowedValues", [])
        self.constraint_description = param_def.get("ConstraintDescription", None)
        self.default = param_def.get("Default", None)
        self.max_length = param_def.get("MaxLength", None)
        self.max_value = param_def.get("MaxValue", None)
        self.min_length = param_def.get("MinLength", None)
        self.min_value = param_def.get("MinValue", None)
        self.no_echo = param_def.get("NoEcho", False)
        self.type = param_def.get("Type", None)

    def to_df(self, name:str) -> pd.DataFrame:
        df = pd.DataFrame({
            "Name": [self.id],
            "Type": [self.type],
            "Description": [self.description],
            "Default": [self.default],
            "MaxLength": [self.max_length],
            "MinLength": [self.min_length],
            "MaxValue": [self.max_value],
            "MinValue": [self.min_value],
            "AllowedPattern": [self.allowed_pattern],
            "AllowedValues": [self.allowed_values],
            "ConstraintDescription": [self.constraint_description],
            "NoEcho": [self.no_echo],
        })

        df.index.name = name

        return df


class CfnOutput(object):

    def __init__(self, output_id:str, output_def:dict) -> None:
        self.id = output_id
        self.description = output_def.get("Description", None)
        self.value = output_def.get("Value", None)
        self.export_name = output_def.get("Export", {}).get("Name", None)

    def to_df(self, name:str) -> pd.DataFrame:
        df = pd.DataFrame({
            "Name": [self.id],
            "Description": [self.description],
            "Value": [self.value],
            "ExportName": [self.export_name],
        })
        df.index.name = name

        return df

class CfnMappings(object):

    def __init__(self, mapping_id:str, mapping_def:dict) -> None:
        self.id = mapping_id
        self.mapping_def = mapping_def

    def to_df(self, name:str) -> pd.DataFrame:
        columns = ["Name", "Item", "Key", "Value"]
        df = pd.DataFrame(columns=columns)
        for item, keyval in self.mapping_def.items():
            for key, val in keyval.items():
                df = pd.concat([
                    df, pd.DataFrame(
                        [[self.id, item, key, val]],
                        columns=columns,
                    )
                ])
        df.index.name = name
        return df


from cfn_docgen import cfn_spec



class CfnTemplate(object):

    cfn_specs = {
        "resource_spec": cfn_spec.CfnSpecification(),
        "creation_policy": cfn_spec.CfnCreationPolicy(),
        "deletion_policy": cfn_spec.CfnDeletionPolicy(),
        "update_policy": cfn_spec.CfnUpdatePolicy(),
        "update_replace_policy": cfn_spec.CfnUpdateReplacePolicy(),
    }

    @classmethod
    def reload_specs(cls):
        spec = cfn_spec.CfnSpecification()
        logger.info(spec.region_name)
        cls.cfn_specs = {
            "resource_spec": cfn_spec.CfnSpecification(),
            "creation_policy": cfn_spec.CfnCreationPolicy(),
            "deletion_policy": cfn_spec.CfnDeletionPolicy(),
            "update_policy": cfn_spec.CfnUpdatePolicy(),
            "update_replace_policy": cfn_spec.CfnUpdateReplacePolicy(),
        }
        logger.info(cls.cfn_specs["resource_spec"].region_name)


    @classmethod
    def get_spec(cls, spec_name:str):
        return cls.cfn_specs[spec_name]
        # return {
        #     "resource_spec": cfn_spec.CfnSpecification(),
        #     "creation_policy": cfn_spec.CfnCreationPolicy(),
        #     "deletion_policy": cfn_spec.CfnDeletionPolicy(),
        #     "update_policy": cfn_spec.CfnUpdatePolicy(),
        #     "update_replace_policy": cfn_spec.CfnUpdateReplacePolicy(),
        # }[spec_name]

    def __init__(self, filepath:str, omit:bool=False) -> None:

        self.reload_specs()
        self.filepath = filepath
        self.body = self.load_template()
        self.is_omit = omit


        self.parameters = self.parse_parameters()
        self.mappings = self.parse_mappings()
        self.resources = self.parse_resources()
        self.outputs = self.parse_outputs()

        # TODOs
        self.description = None
        self.rules = None
        self.conditions = None
        self.transform = None
        self.metadata = None


    @info_logging("parameters")
    def parse_parameters(self) -> List[CfnParameter]:
        logger.info("start process for parameter")
        parameters = []
        for param_id, param_def in tqdm.tqdm(self.body.get("Parameters", {}).items(), desc="parameters"):
            parameters.append(CfnParameter(
                param_id, param_def,
            ))
        logger.info("end process for parameter")
        return parameters

    @info_logging("mappings")
    def parse_mappings(self) -> List[CfnMappings]:
        mappings = []
        for mapping_id, mapping_def in tqdm.tqdm(self.body.get("Mappings", {}).items(), desc="mappings"):
            mappings.append(CfnMappings(
                mapping_id, mapping_def,
            ))
        return mappings


    @info_logging("resources")
    def parse_resources(self) -> List[CfnResource]:
        resources = []
        for resource_id, resource_def in tqdm.tqdm(self.body["Resources"].items(), desc="resources"):
            resource_type = resource_def["Type"]
            if resource_type == "AWS::CDK::Metadata":
                continue
            resources.append(CfnResource(
                resource_id, resource_def,
            ))
        return resources

    @info_logging("outputs")
    def parse_outputs(self) -> List[CfnOutput]:
        outputs = []
        for output_id, output_def in tqdm.tqdm(self.body.get("Outputs", {}).items(), desc="outputs"):
            outputs.append(CfnOutput(
                output_id, output_def,
            ))
        return outputs


    def to_df(self, items:list, name:str) -> pd.DataFrame:

        df = pd.DataFrame()
        all_df = [item.to_df(name) for item in items]

        all_df = [df for df in all_df if df is not None]

        df = pd.concat([df, *all_df])

        # set template level columns
        df["Filename"] = os.path.basename(self.filepath)

        df = df.reset_index(drop=True)
        return df

    def merge_df(self) -> dict:
        targets = {
            "Parameters": self.parameters,
            "Mappings": self.mappings,
            "Resources_Property_Summary": self.resources,
            "Resources_Property_Detail": self.resources,
            "Resources_CreationPolicy": self.resources,
            "Resources_DeletionPolicy": self.resources,
            "Resources_DependsOn": self.resources,
            "Resources_UpdatePolicy": self.resources,
            "Resources_UpdateReplacePolicy": self.resources,
            "Outputs": self.outputs,
        }
        dfs = dict()
        for name, items in targets.items():
            df = self.to_df(items=items, name=name)
            if name == "Resources_Property_Detail":
                df = df.sort_values(["ResourceType", "ResourceId", "Property"])
                if self.is_omit:
                    df = df[df["IsOmittable"] == False]
            dfs[name] = df

        # summarize properties
        target_cols = ["ResourceType", "ResourceId", "ResourceNote"]
        dfs["Resources_Property_Summary"] = (
            dfs["Resources_Property_Detail"].groupby(target_cols, dropna=False)
            .count().reset_index().sort_values(target_cols)
        )[target_cols]
        return dfs

    def load_template(self) -> dict:
        """
        load json or yaml cfn template as dict
        """
        with open(self.filepath, "r") as fp:
            raw = fp.read()
        logger.info(f"load template file from {self.filepath}")
        if self.filepath.split(".")[-1] != "json":
            logger.debug(f"convert template from yaml to json")
            return json.loads(to_json(raw))
        else:
            return json.loads(raw)
    
    def to_file(self, filepath:str=None, fmt:str="xlsx"):
        method_by_fmt = {
            "xlsx": self.to_excel,
            "md": self.to_md,
            "csv": self.to_csv,
            "html": self.to_html,
        }

        if filepath is None:
            filepath = f"{os.path.splitext(self.filepath)[0]}.{fmt}"
        else:
            if fmt != "xlsx":
                filepath = f"{os.path.splitext(self.filepath)[0]}.{fmt}"
            else:
                fmt = filepath.split(".")[-1]
        method_by_fmt[fmt](filepath)

        logger.info(f"save generated file as {fmt} at {filepath}")

    def to_md(self, filepath:str):
        dfs = self.merge_df()

        for name, df in dfs.items():
            path, ext = os.path.splitext(filepath)
            path = f"{path}_{name}{ext}"
            try:
                os.remove(path)
            except FileNotFoundError:
                pass
            df.to_markdown(path, index=False)
        
    def to_csv(self, filepath:str):
        dfs = self.merge_df()
        for name, df in dfs.items():
            path, ext = os.path.splitext(filepath)
            path = f"{path}_{name}{ext}"
            try:
                os.remove(path)
            except FileNotFoundError:
                pass
            df.to_csv(path, index=False)

    def to_html(self, filepath:str):
        dfs = self.merge_df()
        for name, df in dfs.items():
            path, ext = os.path.splitext(filepath)
            path = f"{path}_{name}{ext}"
            try:
                os.remove(path)
            except FileNotFoundError:
                pass
            df.to_html(path, index=False)
        

    def to_excel(self, filepath:str):
        dfs = self.merge_df()
        try:
            os.remove(filepath)
        except FileNotFoundError:
            pass
        with pd.ExcelWriter(filepath, ) as writer:
            wb = writer.book
            for name, df in dfs.items():
                ws = wb.add_worksheet(name)
                # ws = wb.sheets[name]
                (max_row, max_col) = df.shape
                ws.autofilter(1, 1, max_row, max_col - 1)

        style = Styler(horizontal_alignment=utils.horizontal_alignments.left)
        with StyleFrame.ExcelWriter(filepath, mode="a", if_sheet_exists="overlay") as writer:
            for name, df in dfs.items():
                sf = StyleFrame(df)
                sf.set_column_width(columns=sf.columns, width=25)
                if "Description" in sf.columns.tolist():
                    sf.set_column_width(columns="Description", width=100)
                sf.apply_column_style(cols_to_style=sf.columns, styler_obj=style)

                sf.to_excel(
                    writer, index=False, sheet_name=name,
                    startcol=1, startrow=1,
                    freeze_panes=(2,2),
                )



import os
import json
import glob
from typing import List
from cfn_docgen import cfn_spec

def create_template_per_resources(save_dir:str)->List[str]:
    files = glob.glob(os.path.join(save_dir, "*.json"))
    if len(files):
        return files

    spec = cfn_spec.CfnSpecification()
    resource_types = [r for r in spec.spec["ResourceTypes"]]
    empty_template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Description": "AWS CloudFormation Sample Template ElasticBeanstalk_in_VPC: Sample template showing how to create an Elastic Beanstalk environment in a VPC. The stack contains 2 subnets: the first subnet is public and contains the load balancer, a NAT device for internet access from the private subnet and a bastion host to allow SSH access to the Elastic Beanstalk hosts. The second subnet is private and contains the Elastic Beanstalk instances. You will be billed for the AWS resources used if you create a stack from this template.",
        "Resources": [],
    }
    resources = dict()
    for resource_type in resource_types:
        resources[resource_type.replace(":", "")] = {
            "Type": resource_type,
            "Properties": {}
        }

    os.makedirs(save_dir, exist_ok=True)
    for key, val in resources.items():
        path = os.path.join(save_dir, f"{key}.json") 
        empty_template["Resources"] = {key: val}
        with open(path, "w") as fp:
            json.dump(empty_template, fp, indent=4)    

    files = glob.glob(os.path.join(save_dir, "*.json"))
    return files



import json
import unittest

import os
from cfn_docgen import cfn_def

class CfnDefTestCase(unittest.TestCase):

    default_region = "ap-northeast-1"

    def setUp(self) -> None:
        os.environ["CFN_DOCGEN_AWS_REGION"] = self.default_region
        return super().setUp()

    def test_from_json(self):

        resource_id = "VPC"
        resource_def = {
            'Type': 'AWS::EC2::VPC',
            'Metadata': {
                'UserNotes': {
                    'ResourceNote': 'これはVPCリソースに対するユーザ独自のコメントです',
                    'PropNotes': {
                        'CidrBlock': 'これはCidrBlockプロパティに対するユーザ独自のコメントです',
                        'Tags[1].Key': 'これはTagsプロパティ配列の2番目のKeyプロパティに対するユーザ独自のコメントです'
                    }
                }
            },
            'Properties': {
                'CidrBlock': '10.0.0.0/16',
                'Tags': [
                    {'Key': 'ENV', 'Value': 'DEV'},
                    {'Key': 'DEPARTMENT', 'Value': 'SOMU'}
                ]
            }
        }

        resource = cfn_def.CfnResource(resource_id, resource_def)

        resource_json = resource.to_df("Resource_Property_Detail").to_dict(orient="records")
        expected_resource = {
            resource_id: resource_def
        }
        ret_resource = cfn_def.CfnResource.from_json_def(resource_json)
        print(json.dumps(ret_resource, indent=2))
        self.assertDictEqual(ret_resource, expected_resource)



from copy import deepcopy
import os
import requests
import json
from bs4 import BeautifulSoup
import re
import boto3 

from cfn_docgen import util

logger = util.get_module_logger(__name__, util.get_verbose())


class CfnParameter(object):
    def __init__(self) -> None:
        pass
    
    def get_document_url(self):
        return "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/parameters-section-structure.html"

    def get_definition(self):
        return {
            "Description": {
                "Required": False,
                "Type": "String"
            },
            "MaxValue": {
                "Required": False,
                "Type": "Number"
            },
            "MinValue": {
                "Required": False,
                "Type": "Number"
            },
            "MaxLength": {
                "Required": False,
                "Type": "Number"
            },
            "MinLength": {
                "Required": False,
                "Type": "Number"
            },
            "AllowedPattern": {
                "Required": False,
                "Type": "String"
            },
            "AllowedValues": {
                "Required": False,
                "Type": "List"
            },
            "ConstraintDescription": {
                "Required": False,
                "Type": "String"
            },
            "Default": {
                "Required": False,
                "Type": "String"
            },
            "NoEcho": {
                "Required": False,
                "Type": "String"
            },
            "Type": {
                "Required": True,
                "Type": "String"
            }
        }

    def list_allowed_types(self):
        return [
            "String",
            "Number",
            "CommaDelimitedList",
            "List<Number>",
            "AWS::EC2::AvailabilityZone::Name",
            "AWS::SSM::Parameter::Name",
            "AWS::SSM::Parameter::Value<String>",
            "AWS::SSM::Parameter::Value<List<String>>",
            "AWS::SSM::Parameter::Value<CommaDelimitedList>",
            "AWS::EC2::Image::Id",
            "AWS::EC2::Instance::Id",
            "AWS::EC2::KeyPair::KeyName",
            "AWS::EC2::SecurityGroup::GroupName",
            "AWS::EC2::SecurityGroup::Id",
            "AWS::EC2::Subnet::Id",
            "AWS::EC2::VPC::Id",
            "AWS::Route53::HostedZone::Id",
            "List<AWS::EC2::AvailabilityZone::Name>",
            "List<AWS::EC2::Image::Id>",
            "List<AWS::EC2::Instance::Id>",
            "List<AWS::EC2::SecurityGroup::GroupName>",
            "List<AWS::EC2::SecurityGroup::Id>",
            "List<AWS::EC2::Subnet::Id>",
            "AWS::EC2::Volume::Id",
            "List<AWS::EC2::Volume::Id>",
            "List<AWS::EC2::VPC::Id>",
            "List<AWS::Route53::HostedZone::Id>",
            "AWS::SSM::Parameter::Value<AWS::EC2::AvailabilityZone::Name>",
            "AWS::SSM::Parameter::Value<AWS::EC2::Image::Id>",
            "AWS::SSM::Parameter::Value<AWS::EC2::Instance::Id>",
            "AWS::SSM::Parameter::Value<AWS::EC2::KeyPair::KeyName>",
            "AWS::SSM::Parameter::Value<AWS::EC2::SecurityGroup::GroupName>",
            "AWS::SSM::Parameter::Value<AWS::EC2::SecurityGroup::Id>",
            "AWS::SSM::Parameter::Value<AWS::EC2::Subnet::Id>",
            "AWS::SSM::Parameter::Value<AWS::EC2::Volume::Id>",
            "AWS::SSM::Parameter::Value<AWS::EC2::VPC::Id>",
            "AWS::SSM::Parameter::Value<AWS::Route53::HostedZone::Id>",
            "AWS::SSM::Parameter::Value<List<AWS::EC2::AvailabilityZone::Name>>",
            "AWS::SSM::Parameter::Value<List<AWS::EC2::Image::Id>>",
            "AWS::SSM::Parameter::Value<List<AWS::EC2::Instance::Id>>",
            "AWS::SSM::Parameter::Value<List<AWS::EC2::KeyPair::KeyName>>",
            "AWS::SSM::Parameter::Value<List<AWS::EC2::SecurityGroup::GroupName>>",
            "AWS::SSM::Parameter::Value<List<AWS::EC2::SecurityGroup::Id>>",
            "AWS::SSM::Parameter::Value<List<AWS::EC2::Subnet::Id>>",
            "AWS::SSM::Parameter::Value<List<AWS::EC2::Volume::Id>>",
            "AWS::SSM::Parameter::Value<List<AWS::EC2::VPC::Id>>",
            "AWS::SSM::Parameter::Value<List<AWS::Route53::HostedZone::Id>>",
        ]

class CfnResourceAttribute(object):

    doc_url = "https://docs.aws.amazon.com/ja_jp/AWSCloudFormation/latest/UserGuide/aws-product-attribute-reference.html"

    def __init__(self) -> None:
        self.doc = self._load_doc()
        self.allowrd_resources = self.get_allowed_resources()

    def _load_doc(self) -> str:

        local_filepath = os.path.join(util.CACHE_BASE_DIR, os.path.basename(self.doc_url))
        if not os.path.exists(local_filepath):
            resp = requests.get(self.doc_url)
            with open(local_filepath, "w", encoding="utf-8") as fp:
                fp.write(resp.content.decode())
        
        with open(local_filepath, "r", encoding="utf-8") as fp:
            raw = fp.read()
        return raw


    def get_allowed_resources(self) -> list:

        soup = BeautifulSoup(self.doc, "html.parser")

        list_items = soup.find_all("li", {"class": "listitem"})
        resources = []
        for item in list_items:
            resources.append(item.get_text(strip=True))

        return resources

    def get_definition(self):
        raise NotImplementedError


class CfnUpdatePolicy(CfnResourceAttribute):

    doc_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-attribute-updatepolicy.html"

    def __init__(self) -> None:
        self.doc = self._load_doc()
        self.allowed_resources = self.get_allowed_resources()

    def get_definition(self, resource_type:str) -> dict:
        if not resource_type in self.allowed_resources:
            return None
        
        if resource_type == "AWS::AppStream::Fleet":
            return {
                "StopBeforeUpdate": {
                    "Type": {
                        "Type": "Boolean",
                        "Required": "No",
                        "Default": "",
                        "Description": "",
                    },
                    "Description": "Stops the specified fleet before the update."
                },
                "StartAfterUpdate": {
                    "Type": {
                        "Type": "Boolean",
                        "Required": "No",
                        "Default": "",
                        "Description": "",
                    },
                    "Description": "Starts the specified fleet after the update."
                },
            }

        if resource_type == "AWS::AutoScaling::AutoScalingGroup":
            return {
                "AutoScalingReplacingUpdate" : {
                    "WillReplace" : {
                        "Type": "Boolean",
                        "Required": "No",
                        "Defaul": "",
                        "Description": "Specifies whether an Auto Scaling group and the instances it contains are replaced during an update. During replacement, CloudFormation retains the old group until it finishes creating the new one. If the update fails, CloudFormation can roll back to the old Auto Scaling group and delete the new Auto Scaling group. While CloudFormation creates the new group, it doesn't detach or attach any instances. After successfully creating the new Auto Scaling group, CloudFormation deletes the old Auto Scaling group during the cleanup process. When you set the WillReplace parameter, remember to specify a matching CreationPolicy. If the minimum number of instances (specified by the MinSuccessfulInstancesPercent property) don't signal success within the Timeout period (specified in the CreationPolicy policy), the replacement update fails and AWS CloudFormation rolls back to the old Auto Scaling group.",
                    },
                    "Description": "To specify how AWS CloudFormation handles replacement updates for an Auto Scaling group, use the AutoScalingReplacingUpdate policy. This policy enables you to specify whether AWS CloudFormation replaces an Auto Scaling group with a new one or replaces only the instances in the Auto Scaling group.",
                },
                "AutoScalingRollingUpdate" : {
                    "MaxBatchSize" : {
                        "Type": "Integer",
                        "Required": "No",
                        "Defaul": 1,
                        "Description": "Specifies the maximum number of instances that CloudFormation updates.",
                    },
                    "MinInstancesInService" : {
                        "Type": "Integer",
                        "Required": "No",
                        "Defaul": 0,
                        "Description": "Specifies the minimum number of instances that must be in service within the Auto Scaling group while CloudFormation updates old instances. This value must be less than the MaxSize of the Auto Scaling group.",
                    },
                    "MinSuccessfulInstancesPercent" : {
                        "Type": "Integer",
                        "Required": "No",
                        "Defaul": 100,
                        "Description": "Specifies the percentage of instances in an Auto Scaling rolling update that must signal success for an update to succeed. You can specify a value from 0 to 100. CloudFormation rounds to the nearest tenth of a percent. For example, if you update five instances with a minimum successful percentage of 50, three instances must signal success. If an instance doesn't send a signal within the time specified in the PauseTime property, CloudFormation assumes that the instance wasn't updated. If you specify this property, you must also enable the WaitOnResourceSignals and PauseTime properties. The MinSuccessfulInstancesPercent parameter applies only to instances only for signaling purpose. To specify the number of instances in your autoscaling group, see the MinSize, MaxSize, and DesiredCapacity properties for the AWS::AutoScaling::AutoScalingGroup resource.",
                    },
                    "PauseTime" : {
                        "Type": "String",
                        "Required": "No",
                        "Defaul": "PT0S (0 seconds). If the WaitOnResourceSignals property is set to true, the default is PT5M.",
                        "Description": "The amount of time that CloudFormation pauses after making a change to a batch of instances to give those instances time to start software applications. For example, you might need to specify PauseTime when scaling up the number of instances in an Auto Scaling group. If you enable the WaitOnResourceSignals property, PauseTime is the amount of time that CloudFormation should wait for the Auto Scaling group to receive the required number of valid signals from added or replaced instances. If the PauseTime is exceeded before the Auto Scaling group receives the required number of signals, the update fails. For best results, specify a time period that gives your applications sufficient time to get started. If the update needs to be rolled back, a short PauseTime can cause the rollback to fail. Specify PauseTime in the ISO8601 duration format (in the format PT#H#M#S, where each # is the number of hours, minutes, and seconds, respectively). The maximum PauseTime is one hour (PT1H)",
                    },
                    "SuspendProcesses" : {
                        "Type": "List of Auto Scaling processes",
                        "Required": "No",
                        "Defaul": "",
                        "Description": "Specifies the Auto Scaling processes to suspend during a stack update. Suspending processes prevents Auto Scaling from interfering with a stack update. For example, you can suspend alarming so that Amazon EC2 Auto Scaling doesn't initiate scaling policies associated with an alarm. For valid values, see the ScalingProcesses.member.N parameter for the SuspendProcesses action in the Amazon EC2 Auto Scaling API Reference.",
                    },
                    "WaitOnResourceSignals" : {
                        "Type": "Boolean",
                        "Required": "Conditional. If you specify the MinSuccessfulInstancesPercent property, you must also enable the WaitOnResourceSignals and PauseTime properties.",
                        "Defaul": False,
                        "Description": "Specifies whether the Auto Scaling group waits on signals from new instances during an update. Use this property to ensure that instances have completed installing and configuring applications before the Auto Scaling group update proceeds. AWS CloudFormation suspends the update of an Auto Scaling group after new EC2 instances are launched into the group. AWS CloudFormation must receive a signal from each new instance within the specified PauseTime before continuing the update. To signal the Auto Scaling group, use the cfn-signal helper script or SignalResource API. To have instances wait for an Elastic Load Balancing health check before they signal success, add a health-check verification by using the cfn-init helper script. For an example, see the verify_instance_health command in the Auto Scaling rolling updates sample template.",
                    },
                    "Description": "To specify how CloudFormation handles rolling updates for an Auto Scaling group, use the AutoScalingRollingUpdate policy. Rolling updates enable you to specify whether AWS CloudFormation updates instances that are in an Auto Scaling group in batches or all at once.",
                },
                "AutoScalingScheduledAction": {
                    "IgnoreUnmodifiedGroupSizeProperties" : {
                        "Type": "Boolean",
                        "Required": "No",
                        "Defaul": False,
                        "Description": "If true, AWS CloudFormation ignores differences in group size properties between your current Auto Scaling group and the Auto Scaling group described in the AWS::AutoScaling::AutoScalingGroup resource of your template during a stack update. If you modify any of the group size property values in your template, AWS CloudFormation uses the modified values and updates your Auto Scaling group.",
                    },
                    "Description": "To specify how AWS CloudFormation handles updates for the MinSize, MaxSize, and DesiredCapacity properties when the AWS::AutoScaling::AutoScalingGroup resource has an associated scheduled action, use the AutoScalingScheduledAction policy. With scheduled actions, the group size properties of an Auto Scaling group can change at any time. When you update a stack with an Auto Scaling group and scheduled action, CloudFormation always sets the group size property values of your Auto Scaling group to the values that are defined in the AWS::AutoScaling::AutoScalingGroup resource of your template, even if a scheduled action is in effect. If you don't want CloudFormation to change any of the group size property values when you have a scheduled action in effect, use the AutoScalingScheduledAction update policy and set IgnoreUnmodifiedGroupSizeProperties to true to prevent CloudFormation from changing the MinSize, MaxSize, or DesiredCapacity properties unless you have modified these values in your template.",
                },
            }
        
        if resource_type == "AWS::ElastiCache::ReplicationGroup":
            return {
                "UseOnlineResharding": {
                    "UseOnlineResharding": {
                        "Type": "Boolean",
                        "Required": "No",
                        "Default": "",
                        "Description": "To modify a replication group's shards by adding or removing shards, rather than replacing the entire AWS::ElastiCache::ReplicationGroup resource, use the UseOnlineResharding update policy.",
                    },
                    "Description": "To modify a replication group's shards by adding or removing shards, rather than replacing the entire AWS::ElastiCache::ReplicationGroup resource, use the UseOnlineResharding update policy."
                }
            }

        if resource_type in ["AWS::OpenSearchService::Domain", "AWS::Elasticsearch::Domain"]:
            return {
                "EnableVersionUpgrade": {
                    "EnableVersionUpgrade": {
                        "Type": "Boolean",
                        "Required": "No",
                        "Default": "",
                        "Description": "To upgrade an OpenSearch Service domain to a new version of OpenSearch or Elasticsearch rather than replacing the entire AWS::OpenSearchService::Domain or AWS::Elasticsearch::Domain resource, use the EnableVersionUpgrade update policy.",
                    },
                    "Description": "To upgrade an OpenSearch Service domain to a new version of OpenSearch or Elasticsearch rather than replacing the entire AWS::OpenSearchService::Domain or AWS::Elasticsearch::Domain resource, use the EnableVersionUpgrade update policy.",
                }
            }
        
        if resource_type == "AWS::Lambda::Alias":
            return {
                "CodeDeployLambdaAlias": {
                    "Description": "To perform an CodeDeploy deployment when the version changes on an AWS::Lambda::Alias resource, use the CodeDeployLambdaAliasUpdate update policy.",
                    "AfterAllowTrafficHook": {
                        "Type": "String",
                        "Required": "No",
                        "Default": "",
                        "Description": "The name of the Lambda function to run after traffic routing completes.",
                    },
                    "ApplicationName": {
                        "Type": "String",
                        "Required": "Yes",
                        "Default": "",
                        "Description": "The name of the CodeDeploy application.",
                    },
                    "BeforeAllowTrafficHook": {
                        "Type": "String",
                        "Required": "No",
                        "Default": "",
                        "Description": "The name of the Lambda function to run before traffic routing starts.",
                    },
                    "DeploymentGroupName": {
                        "Type": "String",
                        "Required": "No",
                        "Default": "",
                        "Description": "The name of the CodeDeploy deployment group. This is where the traffic-shifting policy is set.",
                    },
                }
            }



class CfnDeletionPolicy(CfnResourceAttribute):

    doc_url = "https://docs.aws.amazon.com/ja_jp/AWSCloudFormation/latest/UserGuide/aws-attribute-deletionpolicy.html"

    def __init__(self) -> None:
        self.doc = self._load_doc()
        self.allowed_resources = self.get_allowed_resources()

    def get_definition(self, resource_type:str) -> dict:
        if not resource_type in self.allowed_resources:
            return None
        
        return {
            "Delete": {
                "Description": "CloudFormation deletes the resource and all its content if applicable during stack deletion. You can add this deletion policy to any resource type. By default, if you don't specify a DeletionPolicy, CloudFormation deletes your resources. However, be aware of the following considerations: For AWS::RDS::DBCluster resources, the default policy is Snapshot. For AWS::RDS::DBInstance resources that don't specify the DBClusterIdentifier property, the default policy is Snapshot. For Amazon S3 buckets, you must delete all objects in the bucket for deletion to succeed.",
            },
            "Retain": {
                "Description": "CloudFormation keeps the resource without deleting the resource or its contents when its stack is deleted. You can add this deletion policy to any resource type. When CloudFormation completes the stack deletion, the stack will be in Delete_Complete state; however, resources that are retained continue to exist and continue to incur applicable charges until you delete those resources. For update operations, the following considerations apply: If a resource is deleted, the DeletionPolicy retains the physical resource but ensures that it's deleted from CloudFormation's scope. If a resource is updated such that a new physical resource is created to replace the old resource, then the old resource is completely deleted, including from CloudFormation's scope.",
            },
            "Snapshot": {
                "Description": "For resources that support snapshots, CloudFormation creates a snapshot for the resource before deleting it. When CloudFormation completes the stack deletion, the stack will be in the Delete_Complete state; however, the snapshots that are created with this policy continue to exist and continue to incur applicable charges until you delete those snapshots.",
            }
        }


class CfnCreationPolicy(CfnResourceAttribute):

    doc_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-attribute-creationpolicy.html"

    def __init__(self) -> None:
        self.doc = self._load_doc()
        self.allowed_resources = self.get_allowed_resources()


    def get_definition(self, resource_type:str) -> dict:
        if not resource_type in self.allowed_resources:
            return None
        
        if resource_type == "AWS::AppStream::Fleet":
            return {
                "StartFleet": {
                    "Type": {
                        "Type": "Boolean",
                        "Required": "No",
                        "Default": "",
                        "Description": "",
                    },
                    "Description": "Starts the specified fleet."
                }
            }

        return {
            "AutoScalingCreationPolicy": {
                "MinSuccessfulInstancesPercent": {
                    "Type": "Integer",
                    "Required": "No",
                    "Default": 100,
                    "Description": "Specifies the percentage of instances in an Auto Scaling replacement update that must signal success for the update to succeed. You can specify a value from 0 to 100. CloudFormation rounds to the nearest tenth of a percent. For example, if you update five instances with a minimum successful percentage of 50, three instances must signal success. If an instance doesn't send a signal within the time specified by the Timeout property, CloudFormation assumes that the instance wasn't created.",
                },
                "Description": "For an Auto Scaling group replacement update, specifies how many instances must signal success for the update to succeed."
            },
            "ResourceSignal": {
                "Count": {
                    "Type": "Integer",
                    "Required": "No",
                    "Default": 1,
                    "Description": "The number of success signals CloudFormation must receive before it sets the resource status as CREATE_COMPLETE. If the resource receives a failure signal or doesn't receive the specified number of signals before the timeout period expires, the resource creation fails and CloudFormation rolls the stack back.",
                },
                "Timeout": {
                    "Type": "String",
                    "Required": "No",
                    "Default": "PT5M",
                    "Description": "The length of time that CloudFormation waits for the number of signals that was specified in the Count property. The timeout period starts after CloudFormation starts creating the resource, and the timeout expires no sooner than the time you specify but can occur shortly thereafter. The maximum time that you can specify is 12 hours. The value must be in ISO8601 duration format, in the form: PT#H#M#S, where each # is the number of hours, minutes, and seconds, respectively. For best results, specify a period of time that gives your instances plenty of time to get up and running. A shorter timeout can cause a rollback.",
                },
                "Description": "When CloudFormation creates the associated resource, configures the number of required success signals and the length of time that CloudFormation waits for those signals."
            }
        }



class CfnUpdateReplacePolicy(CfnResourceAttribute):

    doc_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-attribute-updatereplacepolicy.html"

    def __init__(self) -> None:
        self.doc = self._load_doc()
        self.allowed_resources = self.get_allowed_resources()

    def get_definition(self, resource_type:str) -> dict:
        if not resource_type in self.allowed_resources:
            return None
        
        return {
            "Delete": {
                "Description": "CloudFormation deletes the resource and all its content if applicable during resource replacement. You can add this policy to any resource type. By default, if you don't specify an UpdateReplacePolicy, CloudFormation deletes your resources. However, be aware of the following consideration: For Amazon S3 buckets, you must delete all objects in the bucket for deletion to succeed.",
            },
            "Retain": {
                "Description": "CloudFormation keeps the resource without deleting the resource or its contents when the resource is replaced. You can add this policy to any resource type. Resources that are retained continue to exist and continue to incur applicable charges until you delete those resources. If a resource is replaced, the UpdateReplacePolicy retains the old physical resource but removes it from CloudFormation's scope.",
            },
            "Snapshot": {
                "Description": "For resources that support snapshots, CloudFormation creates a snapshot for the resource before deleting it. Snapshots that are created with this policy continue to exist and continue to incur applicable charges until you delete those snapshots.",
            }
        }



class CfnSpecification(object):

    spec_url_by_region = {
        "us-east-2": "https://dnwj8swjjbsbt.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "us-east-1": "https://d1uauaxba7bl26.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "us-west-1": "https://d68hl49wbnanq.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "us-west-2": "https://d201a2mn26r7lk.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "af-south-1": "https://cfn-resource-specifications-af-south-1-prod.s3.af-south-1.amazonaws.com/latest/gzip/CloudFormationResourceSpecification.json",
        "ap-east-1": "https://cfn-resource-specifications-ap-east-1-prod.s3.ap-east-1.amazonaws.com/latest/gzip/CloudFormationResourceSpecification.json",
        "ap-southeast-3": "https://cfn-resource-specifications-ap-southeast-3-prod.s3.ap-southeast-3.amazonaws.com/latest/CloudFormationResourceSpecification.json",
        "ap-south-1": "https://d2senuesg1djtx.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "ap-northeast-2": "https://d1ane3fvebulky.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "ap-southeast-1": "https://doigdx0kgq9el.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "ap-southeast-2": "https://d2stg8d246z9di.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "ap-northeast-1": "https://d33vqc0rt9ld30.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "ap-northeast-3": "https://d2zq80gdmjim8k.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "ca-central-1": "https://d2s8ygphhesbe7.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "cn-north-1": "https://cfn-resource-specifications-cn-north-1-prod.s3.cn-north-1.amazonaws.com.cn/latest/gzip/CloudFormationResourceSpecification.json",
        "cn-northwest-1": "https://cfn-resource-specifications-cn-northwest-1-prod.s3.cn-northwest-1.amazonaws.com.cn/latest/gzip/CloudFormationResourceSpecification.json",
        "eu-central-1": "https://d1mta8qj7i28i2.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "eu-west-1": "https://d3teyb21fexa9r.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "eu-west-2": "https://d1742qcu2c1ncx.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "eu-west-3": "https://d2d0mfegowb3wk.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "eu-north-1": "https://diy8iv58sj6ba.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "eu-south-1": "https://cfn-resource-specifications-eu-south-1-prod.s3.eu-south-1.amazonaws.com/latest/gzip/CloudFormationResourceSpecification.json",
        "me-south-1": "https://cfn-resource-specifications-me-south-1-prod.s3.me-south-1.amazonaws.com/latest/gzip/CloudFormationResourceSpecification.json",
        "sa-east-1": "https://d3c9jyj3w509b0.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json",
        "us-gov-east-1": "https://s3.us-gov-east-1.amazonaws.com/cfn-resource-specifications-us-gov-east-1-prod/latest/CloudFormationResourceSpecification.json",
        "us-gov-west-1": "https://s3.us-gov-west-1.amazonaws.com/cfn-resource-specifications-us-gov-west-1-prod/latest/CloudFormationResourceSpecification.json",
    }

    spec_cach_by_region = dict()


    def __init__(self, filepath:str=None, region_name:str=None) -> None:
        self.filepath = filepath
        self.region_name = self.get_region_name(region_name)
        self.spec = dict()
        self.spec = self.load_spec()
        # try:
        #     self.spec = self.spec_cach_by_region[self.region_name]
        # except KeyError:
        #     self.spec_cach_by_region[self.region_name] = self.load_spec()
        #     self.spec = self.spec_cach_by_region[self.region_name]

        logger.info(f"spec for {self.region_name} loaded")

        self.html_cache = dict()


    def get_region_name(self, region_name:str):
        """
        if region_name are given, use it.
        if region_name are not given, firstly check the environment variable AWS_REGION,
        secondly, check the aws cli profile
        """
        if region_name:
            logger.debug(f"get region name from argument: {region_name}")
            return region_name
        
        try:
            region_name = os.environ["CFN_DOCGEN_AWS_REGION"]
            logger.debug(f"get region name from environment variable : {region_name}")
            return region_name
        except KeyError:
            profile = boto3.client("sts")
            logger.debug(f"get region name from aws cli default profile")
            return profile.meta.region_name


    def load_spec(self) -> dict:
        """
        load spec file from local file.
        if local file does not specified, default is '~/.cfn-docgen/cache/AWS_REGION/CloudFormationResourceSpecification.json'.
        if local filep does not exists, firstly download it.
        """
        if self.filepath is None:
            cache_dir = os.path.join(util.CACHE_BASE_DIR, self.region_name)
            self.filepath = os.path.join(cache_dir, "CloudFormationResourceSpecification.json")
        
        # make sure the directory already created
        os.makedirs(os.path.dirname(self.filepath), exist_ok=True)
        
        if not os.path.exists(self.filepath):
            url = self.spec_url_by_region[self.region_name]
            resp = requests.get(url)

            with open(self.filepath, "w", encoding="utf-8") as fp:
                fp.write(resp.content.decode())
            logger.debug(f"download cfn spec file from {url}")

        with open(self.filepath, "r", encoding="utf-8") as fp:
            spec =  json.load(fp)
        logger.debug(f"load cfn spec from {self.filepath}")
        return spec


    def get_custom_resource_spec(self, name:str, defs:dict):
        base_spec = self.spec["ResourceTypes"]["AWS::CloudFormation::CustomResource"]
        base_spec = self.get_detailed_resource_spec("AWS::CloudFormation::CustomResource", base_spec)

        custom_spec = deepcopy(base_spec)
        for key in defs["Properties"].keys():
            if key == "ServiceToken":
                continue
            custom_spec["Properties"][key] = {}

        return custom_spec


    def get_resource_spec(self, resource_name:str, resource_def:dict=None) -> dict:
        """
        return a specified resource spec
        """
        try:
            if resource_name.startswith("Custom::") or resource_name == "AWS::CloudFormation::CustomResource":
                custom_resource_spec = self.get_custom_resource_spec(resource_name, resource_def)
                return custom_resource_spec
            
            ret_spec = self.spec["ResourceTypes"][resource_name]
            ret_spec = self.get_detailed_resource_spec(resource_name, ret_spec)
            return ret_spec

        except Exception as ex:
            logger.error(f"get resource spec for {resource_name} failed : {type(ex)} {ex}")
            raise ex
        

    def get_property_spec(self, resource_name:str, property_name:str=None) -> dict:
        """
        return a property spec of specified resource
        property_name, if specified, must be the format like 'AWS::S3::Bucket.CorsRule'
        """
        try:
            if property_name is not None:
                try:
                    ret_spec = {property_name: self.spec["PropertyTypes"][property_name]}
                    # return {property_name: self.spec["PropertyTypes"][property_name]["Properties"]}
                except KeyError:
                    return {property_name: {"Properties": {}}}
            else:
                match_keys = filter(
                    lambda key: key.startswith(f"{resource_name}."),
                    self.spec["PropertyTypes"].keys()
                )
                ret_spec = {key: self.spec["PropertyTypes"][key] for key in match_keys}

            ret_spec = self.get_detailed_prop_spec(resource_name, ret_spec, property_name)
        except Exception as ex:
            logger.error(f"get property spec for {resource_name}/{property_name} failed : {type(ex)} {ex}")
            raise ex
        return ret_spec



    def load_html(self, doc_url:str) -> BeautifulSoup:
        """
        load the html file and return it as BeautifulSoup object
        if the html file does not exist, downalod it from the url
        if the BeautifulSoup object of the html file cached in memory, return it
        """
        try:
            soup = self.html_cache[doc_url]
        except KeyError:
            filepath = os.path.join(
                util.CACHE_BASE_DIR, os.path.basename(doc_url),
            )
            if not os.path.exists(filepath):
                resp = requests.get(doc_url)
                with open(filepath, "w", encoding="utf-8") as fp:
                    fp.write(resp.content.decode())
                logger.debug(f"download cfn reference file from {doc_url}")        
            with open(filepath, "r", encoding="utf-8") as fp:
                doc_body = fp.read()
            soup = BeautifulSoup(doc_body, "html.parser")
            self.html_cache[doc_url] = soup
        return soup

    def is_except_resource(self, resource_type:str, prop_type:str) -> bool:
        if resource_type == "AWS::IoTWireless::TaskDefinition" and prop_type == "LoRaWANUpdateGatewayTaskEntry":
            return True
        if resource_type == "AWS::IoTWireless::TaskDefinition" and prop_type == "TaskDefinitionType":
            return True
        if resource_type == "AWS::OpsWorks::Instance.EbsBlockDevice":
            return True
        if resource_type == "AWS::SageMaker::ModelBiasJobDefinition" and prop_type == "EndpointName":
            return True
        if resource_type == "AWS::SageMaker::ModelQualityJobDefinition" and prop_type == "EndpointName":
            return True
        if resource_type == "AWS::SageMaker::DataQualityJobDefinition" and prop_type == "EndpointName":
            return True
        if resource_type == "AWS::SageMaker::ModelExplainabilityJobDefinition" and prop_type == "EndpointName":
            return True
        if resource_type == "AWS::ApplicationInsights::Application" and prop_type == "GroupingType":
            return True
        if resource_type.startswith("AWS::SageMaker::EndpointConfig.Clarify"):
            return True
        if resource_type == "AWS::ElasticLoadBalancingV2::Listener.AuthenticateOidcConfig" and prop_type == "UseExistingClientSecret":
            return True

    def get_detailed_resource_spec(self, resource_name:str, resource_spec:dict) -> dict:
        """
        add detailed spec information from its CloudFormation User Guide page
        """

        # load html 
        doc_url = resource_spec["Documentation"]
        doc_url = self.modify_doc_url(resource_name, doc_url)
        soup = self.load_html(doc_url)
        
        new_prop_spec = dict()
        prop_id, prop_html_id, = None, None
        try:
            for prop_id, prop in resource_spec["Properties"].items():
                prop_html_id = prop["Documentation"].split("#")[-1]
                try:
                    paragraphs = soup.find("a", {"id": prop_html_id}).parent.next_sibling.find_all("p")
                except AttributeError:
                    try:
                        paragraphs = soup.find("a", {"id": prop_html_id}).parent.next_sibling.next_sibling.find_all("p")
                    except AttributeError as ex:
                        # if self.is_except_resource(resource_name, prop_id):
                        #     continue
                        # else:
                        #     raise ex
                        fail_msg = f"get detailed resource spec for {resource_name}/{prop_id}/{prop_html_id} from {doc_url} failed"
                        logger.warning(fail_msg)
                        prop["DocDescription"] = fail_msg
                        paragraphs = ""
                for i, p in enumerate(paragraphs):
                    # get description
                    if i == 0:
                        description = p.get_text()
                        description = re.sub(r"\s\s+", " ", description.replace("\n", ""))
                        key, value = "Description", description
                    else:
                        try:
                            key, value = p.get_text().split(": ")
                        except ValueError:
                            continue
                    prop[f"Doc{key}"] = value
                new_prop_spec[prop_id] = prop
            resource_spec["Properties"] = new_prop_spec
        except Exception as ex:
            logger.error(f"get detailed resource spec for {resource_name}/{prop_id}/{prop_html_id} from {doc_url} failed : {type(ex)} {ex}")
            raise ex
        
        return resource_spec


    def modify_doc_url(self, property_type:str, doc_url:str) ->str:
        prev_url = doc_url
        if property_type == "AWS::EC2::SecurityGroup.Ingress":
            doc_url = doc_url.replace(".html", "-1.html")

        if property_type == "AWS::AutoScaling::AutoScalingGroup.LaunchTemplateOverrides":
            doc_url = doc_url.replace(
                "cfn-as-mixedinstancespolicy-launchtemplateoverrides.html",
                "aws-properties-autoscaling-autoscalinggroup-launchtemplateoverrides.html",
            )
        if property_type == "AWS::AutoScaling::AutoScalingGroup.InstancesDistribution":
            doc_url = doc_url.replace(
                "cfn-as-mixedinstancespolicy-instancesdistribution.html",
                "aws-properties-autoscaling-autoscalinggroup-instancesdistribution.html",
            )
        if property_type == "AWS::AutoScaling::AutoScalingGroup.InstanceRequirements":
            doc_url = doc_url.replace(
                "cfn-as-mixedinstancespolicy-instancerequirements.html",
                "aws-properties-autoscaling-autoscalinggroup-instancerequirements.html",
            )
        if property_type == "AWS::AutoScaling::AutoScalingGroup.MixedInstancesPolicy":
            doc_url = doc_url.replace(
                "cfn-as-group-mixedinstancespolicy.html",
                "aws-properties-autoscaling-autoscalinggroup-mixedinstancespolicy.html",
            )
        if property_type == "AWS::AutoScaling::AutoScalingGroup.LaunchTemplate":
            doc_url = doc_url.replace(
                "cfn-as-mixedinstancespolicy-launchtemplate.html",
                "aws-properties-autoscaling-autoscalinggroup-launchtemplate.html",
            )

        if property_type == "AWS::RDS::DBParameterGroup":
            doc_url = doc_url.replace(
                "aws-properties-rds-dbparametergroup.html",
                "aws-resource-rds-dbparametergroup.html"
            )

        if property_type == "AWS::EC2::CustomerGateway":
            doc_url = doc_url.replace(
                "aws-resource-ec2-customergateway.html",
                "aws-resource-ec2-customer-gateway.html",
            )

        if prev_url != doc_url:
            logger.debug(f"modify doc url from {prev_url} to {doc_url}")
        return doc_url


    def modify_html_id(self, html_id:str, resource_type:str=None) -> str:
        prev_html_id = html_id
        if html_id == "cfn-as-group-launchtemplate":
            html_id = "cfn-as-group-launchtemplate-launchtemplatespecification"
        elif html_id == "cfn-as-mixedinstancespolicy-overrides":
            html_id = "cfn-as-group-launchtemplate-overrides"
        elif html_id == "cfn-route53-recordset-geolocation-continentcode":
            html_id = "cfn-route53-recordsetgroup-geolocation-continentcode"

        if resource_type is not None and resource_type != "AWS::WAF::ByteMatchSet":

            if html_id == "cfn-waf-sizeconstraintset-sizeconstraint-fieldtomatch-data":
                html_id = "cfn-waf-bytematchset-bytematchtuples-fieldtomatch-data"
            elif html_id == "cfn-waf-sizeconstraintset-sizeconstraint-fieldtomatch-type":
                html_id = "cfn-waf-bytematchset-bytematchtuples-fieldtomatch-type"
            elif html_id == "cfn-waf-bytematchset-bytematchtuples-fieldtomatch-data":
                html_id = "cfn-waf-sizeconstraintset-sizeconstraint-fieldtomatch-data"
            elif html_id == "cfn-waf-bytematchset-bytematchtuples-fieldtomatch-type":
                html_id = "cfn-waf-sizeconstraintset-sizeconstraint-fieldtomatch-type"



        elif html_id == "cfn-opsworks-custcookbooksource-pw":
            html_id = "cfn-opsworks-custcookbooksource-password"

        if prev_html_id != html_id:
            logger.debug(f"modify html id from {prev_html_id} to {html_id}")
        return html_id


    def get_detailed_prop_spec(self, resource_name:str, prop_spec:dict, property_name:str=None) -> dict:
        """
        add detailed spec information from cfn reference pages
        """
        new_prop_spec = dict()
        property_type, prop_id, prop_html_id = None, None, None
        for property_type, properties in prop_spec.items():

            try:
                doc_url = properties["Documentation"]
                doc_url = self.modify_doc_url(property_type, doc_url)

                soup = self.load_html(doc_url)

                if properties.get("Properties") is None:
                    prop = deepcopy(properties)
                    prop_id = property_type.split(".")[-1]
                    # prop["Properties"] = dict()
                    new_prop = dict()
                    new_prop["Properties"] = dict()
                    prop_html_id = os.path.basename(prop["Documentation"]).split(".")[0]
                    prop_html_id = self.modify_html_id(prop_html_id, resource_name)

                    try:
                        paragraphs = soup.find("h1", {"id": prop_html_id}).next_sibling.next_siblings
                    except AttributeError as ex:
                        # if self.is_except_resource(property_type, prop_id):
                        #     continue
                        # else:
                        #     logger.error(f"get detailed property spec for {resource_name}/{property_name}/{property_type}/{prop_id}/{prop_html_id} : {type(ex)} {ex}")
                        #     raise ex
                        fail_msg = f"get detailed property spec for {resource_name}/{property_name}/{property_type}/{prop_id}/{prop_html_id} from {doc_url} failed"
                        logger.warning(fail_msg)
                        prop["DocDescription"] = fail_msg
                        paragraphs = ""

                    for i, p in enumerate(paragraphs):
                        # get description
                        if i == 0:
                            description = p.get_text()
                            description = re.sub(r"\s\s+", " ", description.replace("\n", ""))
                            key, value = "Description", description
                        else:
                            try:
                                text = p.get_text()
                                key, value = text.split(": ")
                            except ValueError as ex:
                                additional_description = re.sub(r"\s\s+", " ", text.replace("\n", ""))
                                key, value = "Description", prop.get("DocDescription", "") + "\n" + additional_description
                                pass
                        prop[f"Doc{key}"] = value
                    new_prop["Properties"][prop_id] = prop
                    new_prop_spec[property_type] = deepcopy(new_prop)
                else:

                    for prop_id, prop in properties["Properties"].items():
                        prop_html_id = prop["Documentation"].split("#")[-1]
                        prop_html_id = self.modify_html_id(prop_html_id, resource_name)

                        # logger.debug(f"{property_type}/{prop_id}")

                        try:
                            paragraphs = soup.find("a", {"id": prop_html_id}).parent.next_sibling.find_all("p")
                        except AttributeError:
                            try:
                                paragraphs = soup.find("a", {"id": prop_html_id}).parent.next_sibling.next_sibling.find_all("p")
                            except AttributeError as ex:
                                # if self.is_except_resource(property_type, prop_id):
                                #     continue
                                # else:
                                #     logger.error(f"get detailed property spec for {resource_name}/{property_name}/{property_type}/{prop_id}/{prop_html_id} : {type(ex)} {ex}")
                                #     raise ex
                                fail_msg = f"get detailed property spec for {resource_name}/{property_name}/{property_type}/{prop_id}/{prop_html_id} from {doc_url} failed"
                                logger.warning(fail_msg)
                                prop["DocDescription"] = fail_msg
                                paragraphs = ""
                        for i, p in enumerate(paragraphs):
                            # get description
                            if i == 0:
                                description = p.get_text()
                                description = re.sub(r"\s\s+", " ", description.replace("\n", ""))
                                key, value = "Description", description
                            else:
                                try:
                                    text = p.get_text()
                                    key, value = text.split(": ")
                                except ValueError as ex:
                                    # logger.debug(f"a skippable error occured {resource_name}/{property_name}/{property_type}/{prop_id} : {type(ex)} {ex}")
                                    additional_description = re.sub(r"\s\s+", " ", text.replace("\n", ""))
                                    key, value = "Description", prop.get("DocDescription", "") + "\n" + additional_description
                                    pass
                            prop[f"Doc{key}"] = value
                        properties["Properties"][prop_id] = prop

                    
                    new_prop_spec[property_type] = properties
            except Exception as ex:
                logger.error(f"get detailed property spec for {resource_name}/{property_name}/{property_type}/{prop_id}/{prop_html_id} from {doc_url} failed : {type(ex)} {ex}")
                raise ex
            
        return new_prop_spec





import os
import click

from cfn_docgen import util

@click.command(name="cfn-docgen")
@click.option("--in", "in_filepath", required=True, help="Input cfn template file path (yaml/json)")
# @click.option("--out", "out_filepath", required=False, default=None, help="Output file path.")
@click.option("--fmt", "fmt", required=False, default="xlsx", help="Output file format.", type=click.Choice(["xlsx", "md", "csv", "html"]), show_default=True)
@click.option("--omit", required=False, default=False, is_flag=True, help="If set, optional properties whose actual values are not set in input template file will not be written in output file.",)
@click.option("--refresh", required=False, default=False, is_flag=True, help="If set, fristly remove all existing cache files and download them again.",)
@click.option("--region", required=False, default=None, help="AWS region name for referencing resource specs. If not set, the value set as environment variable `CFN_DOCGEN_AWS_REGION` is used. If the environment variable is not set, use the value of AWS CLI default profile")
@click.option("--verbose", required=False, is_flag=True, help="If set, stdout DEBUG level logs")
def main(
    in_filepath:str, fmt="xlsx", omit:bool=False, refresh:bool=False, region:str=None, verbose:bool=False,
):
    """
    Document generator from cfn template files
    """

    os.environ["CFN_DOCGEN_VERBOSE"] = str(verbose).upper()
    logger = util.get_module_logger(__name__, verbose=verbose)
    logger.info("start process")

    if refresh:
        util.remove_cache()
        logger.info(f"remove all caches at {util.CACHE_BASE_DIR}")
    
    
    if region is not None:
        os.environ["CFN_DOCGEN_AWS_REGION"] = region



    from cfn_docgen import cfn_def
    # cfn_def.CfnTemplate.reload_specs()
    template = cfn_def.CfnTemplate(
        filepath=in_filepath, omit=omit,
    )
    template.to_file(None, fmt)

    logger.info("end propcess")






import os
from setuptools import setup, find_packages
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = fh.read()

setup(
    name = 'cfn-docgen',
    version = "0.0.17",
    author = 'Takehiro Horie',
    author_email = 'horie.takehiro@outlook.jp',
    license = 'MIT License',
    description = 'Document generator from cfn template files.',
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = 'https://github.com/horietakehiro/cfn-docgen',
    py_modules = ['cfn_docgen'],
    packages = find_packages(),
    install_requires = [requirements],
    python_requires='>=3.8',
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
    ],
    entry_points = '''
        [console_scripts]
        cfn-docgen=cfn_docgen.main:main
    '''
)


import glob
import shutil
import os
from logging import Handler, Logger, getLogger, Formatter, StreamHandler, FileHandler, DEBUG, INFO

APP_BASE_DIR=os.path.join(os.environ["HOME"], ".cfn-docgen")
CACHE_BASE_DIR=os.path.join(APP_BASE_DIR, "cache")
LOG_BASE_DIR=os.path.join(APP_BASE_DIR, "log")
os.makedirs(CACHE_BASE_DIR, exist_ok=True)
os.makedirs(LOG_BASE_DIR, exist_ok=True)

def remove_cache():
    shutil.rmtree(CACHE_BASE_DIR)

def get_verbose():
    verbose = os.environ.get("CFN_DOCGEN_VERBOSE", "FALSE")
    return verbose == "TRUE"

def get_module_logger(module:str, verbose:bool) -> Logger:
    logger = getLogger(module)
    logger = _set_handler(logger, StreamHandler(), verbose)
    logger = _set_handler(logger, FileHandler(os.path.join(LOG_BASE_DIR, "cfn-docgen.log")), True)
    logger.setLevel(DEBUG)
    logger.propagate = False
    return logger

def _set_handler(logger:Logger, handler:Handler, verbose:bool) -> Logger:
    if verbose:
        handler.setLevel(DEBUG)
    else:
        handler.setLevel(INFO)
    handler.setFormatter(Formatter('%(asctime)s %(name)s:%(lineno)s %(funcName)s [%(levelname)s]: %(message)s'))
    logger.addHandler(handler)
    return logger